module ftp_shared
open framework

// shared elements of FTP system

abstract sig Packet {}
sig Block, Hash extends Packet {}

sig File {blocks: set Block, hash: Hash}

fact Hashing {
	-- hashing based solely on blocks
	all f, f': File | f.hash = f'.hash iff f.blocks = f'.blocks
	}

sig Network extends Domain {inpackets, outpackets: set Packet} {
	-- network never contrives hashes for files
	all h: Hash & outpackets | h in inpackets or no f: File | f.hash = h
	-- define reliable network
	this in OK iff inpackets = outpackets
	}

sig FileSystem extends Machine {file: File, client: Client} {
	-- API: file system reads and writes blocks, and delivers hash correctly
	this in OK iff (client.hash = file.hash and client.blocks = file.blocks)
	}

abstract sig Client extends Machine {hash: Hash, blocks: set Block, network: Network}
abstract sig Sender, Receiver extends Client {}

sig FTP_Requirement extends Requirement {
	from, to: FileSystem, sender: Sender, receiver: Receiver, network: Network
	}{
	from != to and no from.file & to.file
	sender = from.client and receiver = to.client
	network = sender.network and network = receiver.network
	}
